"""Constants used in `tools`.

:copyright: (c) 2022-present Tanner B. Corcoran
:license: MIT, see LICENSE for more details.
"""

__author__ = "Tanner B. Corcoran"
__license__ = "MIT License"
__copyright__ = "Copyright (c) 2022-present Tanner B. Corcoran"

ALLOWCHARS = set("abcdefghijklmnopqrstuvwxyz1234567890_")

EMBED_COLOR__DEF = 0xeb9b2a
EMBED_COLOR__POS = 0x43b582
EMBED_COLOR__NEG = 0xeb4747

CREATOR_REFERENCE = "https://github.com/tanrbobanr/dpy-devtools"

M_COMMAND_LOCKED = "This command is currently locked."
